"""
Mimic:
Let Edison mimic your voice in a funny way

"""

import pyaudio
import audioop
import numpy as np
import struct
import wave
import time
import mraa

led_pins=[13,12,11,10]
btn_pins=[9,8,7,6]
levels=[-7,-3,3,7]
leds=[]
btns=[]

CHUNK = 1024
WIDTH = 2
CHANNELS = 1 
RATE = 44100
RECORD_SECONDS = 5
level=-1
swidth=2

#Volume needed to activate the programme
activate_volume=2000

def dimAllLeds():
    for i in range(4):
        leds[i].write(0)

def speedx(snd_array, factor):
    """ Speeds up / slows down a sound, by some factor. """
    indices = np.round( np.arange(0, len(snd_array), factor) )
    indices = indices[indices < len(snd_array)].astype(int)
    return snd_array[ indices ]

def stretch(snd_array, factor, window_size, h):
    """ Stretches/shortens a sound, by some factor. """
    phase  = np.zeros(window_size)
    hanning_window = np.hanning(window_size)
    result = np.zeros( len(snd_array) /factor + window_size)

    for i in np.arange(0, len(snd_array)-(window_size+h), h*factor):

        # two potentially overlapping subarrays
        a1 = snd_array[i: i + window_size]
        a2 = snd_array[i + h: i + window_size + h]

        # the spectra of these arrays
        s1 =  np.fft.fft(hanning_window * a1)
        s2 =  np.fft.fft(hanning_window * a2)

        #  rephase all frequencies
        phase = (phase + np.angle(s2/s1)) % 2*np.pi

        a2_rephased = np.fft.ifft(np.abs(s2)*np.exp(1j*phase))
        i2 = int(i/factor)
        result[i2 : i2 + window_size] += hanning_window*a2_rephased

    result = ((2**(16-4)) * result/result.max()) # normalize (16bit)

    return result.astype('int16')

def pitchshift(snd_array, n, window_size=2**13, h=2**11):
    """ Changes the pitch of a sound by ``n`` semitones. """
    factor = 2**(1.0 * n / 12.0)
    stretched = stretch(snd_array, 1.0/factor, window_size, h)
    return speedx(stretched[window_size:], factor)


if __name__=="__main__":
    #Initialize the buttons and leds
    for i in range(0,4):
        led=mraa.Gpio(led_pins[i])
        led.dir(mraa.DIR_OUT)
        btn=mraa.Gpio(btn_pins[i])
        btn.dir(mraa.DIR_IN)
        btn.mode(mraa.MODE_PULLUP)
        leds.append(led)
        btns.append(btn)

    #Initialize audio stream
    p = pyaudio.PyAudio()
    stream = p.open(format=p.get_format_from_width(WIDTH),
                    channels=CHANNELS,
                    rate=RATE,
                    input=True,
                    output=True,
                    frames_per_buffer=CHUNK)



    #The main loop
    while True:
        frames=""
        stream.start_stream()
        
        #Work to do before mimic starts
        while True:
            #Check if buttons are pressed
            for i in range(4):
                if btns[i].read()==0:
                    dimAllLeds()
                    leds[i].write(1)
                    level=levels[i]
            #Check if someone is speaking 
            data = stream.read(CHUNK)
            rms=audioop.rms(data,2)
            if rms>activate_volume:
                frames=frames+data
                break
                
        #Start recording
        timer=0
        print("* recording")
        while True:
            try:
                data = stream.read(CHUNK)
                frames=frames+data
                rms=audioop.rms(data,2)
                if rms<activate_volume:
                    if time.time()-timer>1:
                        break
                else:
                    timer=time.time()
            except:
                pass
                
        #Do the pitch shifting magic
        frames = struct.unpack("%dh"%(len(frames)/swidth), frames)
        frames = np.array(frames)
        frames=pitchshift(frames, level)
        frames=struct.pack("%dh"%(len(frames)),*list(frames))
        
        #Play the modified sound
        data = frames[0:CHUNK] 
        i=0
        while data != '':
            stream.write(data)
            i=i+1
            data = frames[i*CHUNK:(i+1)*CHUNK] 

        print("* done")
        stream.stop_stream()
        
    stream.close()
    p.terminate()
